/* ***********************************************************************************
	Writer:		Sebastien Bloc
	Copyright:	2004
	eMail:		sebastien.bloc@free.fr
	URL:		http://mignonsoft.free.fr

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	http://www.gnu.org/copyleft/gpl.html
*********************************************************************************** */

/*! \file MyTree.h
\brief Arbre generique
*/

#ifndef INCLUDE_MY_TREE_INCLUDE
#define INCLUDE_MY_TREE_INCLUDE

#include "MyList.h"

/*!
\brief Gestion des Noeud d'un arbre MyTree

Compos par:
	- elem (T)
	- parent (MyTreeNode *)
	- childs (MyList<MyTreeNode *)
*/
template <class T> 
class MyTreeNode
{
public:
	T *elem; 							// element de la liste 
	MyTreeNode *parent;
	MyList<MyTreeNode *> childs;      // enfants 

private:
	void Init()
	{
		elem = NULL;
		parent=NULL;
	}

public:
	MyTreeNode() { Init(); }

	MyTreeNode(T &_elem) 
	{
		Init();
		elem = new T;
		if (!elem) throw MyExceptionMemory("MyTreeNode","constructor",sizeof(T));
		*elem=_elem; //operator=(T &source) de la class appel
	}

	~MyTreeNode() 
	{ 
		if (elem) delete elem; 
	}

	MyTreeNode<T> * operator [] (int num)
	{
		return childs[num];
	}

	T * Elems(int num)
	{
		return childs[num]->elem;
	}

	void operator=(MyTreeNode &source)
	{
		if (source.elem) 
		{
			elem = new T;
			if (!elem) throw CMyExceptionMemory("MyTreeNode","operator=",sizeof(T));
			*elem=*source.elem;
		}
		else elem=NULL;
		parent=source.parent;
		childs=source.childs;
	}
};

/*!
\brief Arbre generique

Parametres:
	- m_root (MyTreeNode *) racine
	- m_current (MyTreeNode *) element en cours (cf voir GoTo)

Ex.: Affiche le Modele suivant dans une Pop-up
	- centre
		- gauche
			- gauche gauche
			- gauche droite
		- droite
			- droite gauche
			- droite droite 

\code
MyString message;
void viewTree(MyTreeNode<MyString> *node)
{
	static int level=0;
	for (int i=0;i<level;i++) message+="| ";
	level++;
	message<<*node->elem<<" ("<<node->childs.GetNbElem()<<" childs)\n";
	for (node->childs=0;node->childs.More();node->childs.GoNext()) 
	viewTree(node->childs.GetElem());
	level--;
}

void main()
{
	MyTree<MyString> tree;
	MyTreeNode<MyString> *niv0,*niv1;
	MyString str;
	str="centre";niv0 = tree.Add(str);
	str="gauche";niv1 = tree.Add(str,niv0);
	str="gauche gauche";tree.Add(str,niv1);
	str="gauche droite";tree.Add(str,niv1);
	str="droite";niv1 = tree.Add(str,niv0);
	str="droite gauche";tree.Add(str,niv1);
	str="droite droite";tree.Add(str,niv1);

	viewTree(tree.GetRoot()); // affichage
	MessageBox(NULL,message,"Arb",0);
}
\endcode
*/

template <class T> 
class MyTree
{
private:
	MyTreeNode<T> *m_root;
	MyTreeNode<T> *m_current;

private:
	void Init()
	{
		m_root = NULL;
		m_current = NULL;
	}

public:
	MyTreeNode<T> * GetRoot() { return m_root; }
	MyTreeNode<T> * GetCurrent() { return m_current; }

	MyTree()
	{
		Init();
	}

	~MyTree()
	{
		SuprAll();
	}

	BOOL SuprAll()
	{
		Init();
		return SuprNode(m_root,TRUE);
	}

	BOOL SuprNode(MyTreeNode<T> *node,BOOL recursif=FALSE)
	{
		if (!node) return FALSE;
		if (node->parent) // s'il y a un parent detruir son fils en question
		{
			BOOL isFind = FALSE;
			for (node->parent->childs=0;node->parent->childs.More();node->parent->childs++)
			{
				if (node->parent->childs.GetElem()==node) // j'ai trouv le fils chez le pere
				{
					node->parent->childs.SuprCurrent();
					isFind=TRUE;
					break;
				}
			}
			if (!isFind) // y'a un pb, un pere mais pas de fils dans ce pere !
				throw MyExceptionMemory("MyTree","SuprNode",sizeof(MyTreeNode<T>));
		}
		if (recursif) // en recursif: destruction de tout ces fils
		{
			for (node->childs=0;node->childs.More();node->childs++)
				if (!SuprNode(node->childs.GetElem(),TRUE)) return FALSE;
		}
		delete node;
		return TRUE;
	}

	MyTreeNode<T> *GoToChild(int numChild,MyTreeNode<T> *parent=NULL)
	{
		MyTreeNode<T> *current = GetChild(numChild,parent);
		if (current) m_current=current;
		return current;
	}

	MyTreeNode<T> * GoTo(MyTreeNode<T> *node)
	{
		MyTreeNode<T> * oldCurrent;
		oldCurrent = m_current;
		m_current=node;
		return oldCurrent;
	}

	MyTreeNode<T> *GetChild(int numChild,MyTreeNode<T> *parent=NULL)
	{
		if (!parent) 
		{
			parent = m_current;
			if (!parent) return NULL;
		}
		return parent->childs[numChild];
	}

	MyTreeNode<T> *GoToParent(MyTreeNode<T> *parent=NULL)
	{
		MyTreeNode<T> *current = GetParent(parent);
		if (current) m_current=current;
		return current;
	}

	MyTreeNode<T> *GetParent(MyTreeNode<T> *parent=NULL)
	{
		if (!parent) 
		{
			parent = m_current;
			if (!parent) return NULL;
		}
		return parent->parent;
	}

	MyTreeNode<T> * Add(T& elem,MyTreeNode<T> *parent=NULL)
	{
		MyTreeNode<T> *node = new MyTreeNode<T>(elem); // Creation d'un nouveau noeud
		if (!parent) parent = m_current;    // Attacher ce noeuf a son parent
		if (parent)                      // si premier alors devient root
		{
			node->parent = parent;
			parent->childs+=node;
		}
		else { m_current = m_root = node; }
		return node;
	}
};

#endif